//*******************************************************************************
//* Copyright (c) 2008-2014 Synchrotron SOLEIL
//* All rights reserved. This program and the accompanying materials
//* are made available under the terms of the GNU Lesser Public License v3
//* which accompanies this distribution, and is available at
//* http://www.gnu.org/licenses/lgpl.html
//******************************************************************************
static const char *RcsId = "$Header: /users/chaize/newsvn/cvsroot/Instrumentation/AcquireWaveformLecroy/src/AcquireWaveformLecroy.cpp,v 1.15 2012-03-09 10:22:12 xavela Exp $";
//+=============================================================================
//
// file :         AcquireWaveformLecroy.cpp
//
// description :  C++ source for the AcquireWaveformLecroy and its commands.
//                The class is derived from Device. It represents the
//                CORBA servant object which will be accessed from the
//                network. All commands which can be executed on the
//                AcquireWaveformLecroy are implemented in this file.
//
// project :      TANGO Device Server
//
// $Author: xavela $
//
// $Revision: 1.15 $
//
// $Log: not supported by cvs2svn $
// Revision 1.14  2012/02/28 13:28:25  xavela
// WaveForm.cpp : ptr which receives data, allocated statically !
//
// Revision 1.13  2010/03/26 09:11:31  vince_soleil
// "Migration_Tango7_Part2"
//
// Revision 1.12  2008/04/24 14:59:23  xavela
// xavier :
// - SocketLecroy file splitted for Win32 and Linux like platforms
// - Bugs fixed in SocketLecroy
//
// Revision 1.11  2008/04/24 13:08:44  syldup
// no message
//
// Revision 1.10  2008/02/29 15:00:12  xavela
// xavier:
// sleep added after a write
//
// Revision 1.9  2008/02/28 09:02:05  xavela
// xavier :
// StateMachine file added !! (oups)
//
// Revision 1.7  2004/10/22 13:24:47  buteau
// Second code reread. Minor modifications
//
// Revision 1.6  2004/10/18 13:17:49  xavela
// xavier : add an attr waveArray2.
// Sorts attr in expert and operator view with Pogo.
// Inconsistencies removed.
//
// Revision 1.5  2004/10/18 09:01:40  buteau
// First code reread
//
// Revision 1.4  2004/10/07 09:21:18  xavela
// xavier : no cout now.
//
// Revision 1.3  2004/10/05 13:11:11  xavela
// Xavier : previous commit failed.
//
// Revision 1.2  2004/10/05 13:09:10  xavela
// Xavier : memory allocation in read_attr_hardware() method.
// calculation of offset of the begining of the struct, which contains the context of the waveform acquisition, in the received block data.
//
// Revision 1.1.1.1  2004/09/29 15:42:14  syldup
// Initial import.
//
//
// copyleft :   Synchrotron SOLEIL
//      L'Orme des Merisiers
//      Saint-Aubin - BP 48

//
//-=============================================================================
//
//      This file is generated by POGO
//  (Program Obviously used to Generate tango Object)
//
//         (c) - Software Engineering Group - ESRF
//=============================================================================



//===================================================================
//
//	The following table gives the correspondence
//	between commands and method name.
//
//  Command name|  Method name
//	----------------------------------------
//  State      |  dev_state()
//  Status     |  dev_status()
//  WriteRead  |  write_read()
//
//===================================================================

#include <AcquireWaveformLecroy.h>
#include <AcquireWaveformLecroyClass.h>
#include <yat/time/Timer.h>

static const std::size_t MAX_STRING_LENGTH  = 1024;

namespace AcquireWaveformLecroy_ns
{
//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::AcquireWaveformLecroy(string &s)
//
// description :  constructor for simulated AcquireWaveformLecroy
//
// in : - cl : Pointer to the DeviceClass object
//      - s : Device name
//
//-----------------------------------------------------------------------------
AcquireWaveformLecroy::AcquireWaveformLecroy(Tango::DeviceClass *cl,string &s)
  :Tango::Device_4Impl(cl,s.c_str())
{
  init_device();
}

AcquireWaveformLecroy::AcquireWaveformLecroy(Tango::DeviceClass *cl,const char *s)
  :Tango::Device_4Impl(cl,s)
{
  init_device();
}

AcquireWaveformLecroy::AcquireWaveformLecroy(Tango::DeviceClass *cl,const char *s,const char *d)
  :Tango::Device_4Impl(cl,s,d)
{
  init_device();
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::delete_device()
//
// description :  will be called at device destruction or at init command.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::delete_device()
{
  //- Delete device's allocated object
  if (attr_rawWaveformData_read) {
   delete [] attr_rawWaveformData_read;
   attr_rawWaveformData_read = 0;
  }
  if (attr_verticalScaledData_read) {
   delete [] attr_verticalScaledData_read;
   attr_verticalScaledData_read = 0;
  }
  if (attr_waveArray1_read) {
    delete attr_waveArray1_read;
    attr_waveArray1_read = 0;
  }
  if (attr_waveArray2_read) {
    delete attr_waveArray2_read;
    attr_waveArray2_read = 0;
  }
  if (attr_waveArrayCount_read) {
    delete attr_waveArrayCount_read;
    attr_waveArrayCount_read = 0;
  }
  if (attr_nominalBits_read) {
    delete attr_nominalBits_read;
    attr_nominalBits_read = 0;
  }
  if (attr_horizontalInterval_read) {
    delete attr_horizontalInterval_read;
    attr_horizontalInterval_read = 0;
  }
  if (attr_horizontalOffset_read) {
    delete attr_horizontalOffset_read;
    attr_horizontalOffset_read = 0;
  }
  if (attr_verticalGain_read) {
    delete attr_verticalGain_read;
    attr_verticalGain_read = 0;
  }
  if (attr_verticalOffset_read) {
    delete attr_verticalOffset_read;
    attr_verticalOffset_read = 0;
  }
  if (attr_triggerTime_read) {
    delete attr_triggerTime_read;
    attr_triggerTime_read = 0;
  }
  
  //- delete the Waveform_Data obj
  if (waveform_mgr_)
  {
    //- ask the task to quit
     this->waveform_mgr_->exit();
    //- !!!!! NEVER TRY TO <delete> a yat4tango::DeviceTask, it commits suicide
    //- upon return of its main function (i.e. entry point)!!!!!!
     this->waveform_mgr_ = 0;
  }
  
/*std::cout << "removing log" << std::endl;
	// DEBUG_STREAM << "Remove the InnerAppender." << endl;
	yat4tango::InnerAppender::release(this);

	// DEBUG_STREAM << "Remove the YatLogAdapter." << endl;
	yat4tango::YatLogAdapter::release();
	
std::cout << "removed log : " << log_deleted << std::endl;*/
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::init_device()
//
// description :  will be called at device initialization.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::init_device()
{
  INFO_STREAM << "AcquireWaveformLecroy::AcquireWaveformLecroy() create device " << device_name << endl;

  //- create read attributes
  //--------------------------------------------
  attr_rawWaveformData_read = new short[MAX_WAVEFORM_DATA_LENGTH];

  attr_verticalScaledData_read =  new double[MAX_WAVEFORM_DATA_LENGTH];

  attr_waveArray1_read = new Tango::DevLong;
  *attr_waveArray1_read = 0;

  attr_waveArray2_read = new Tango::DevLong;
  *attr_waveArray2_read = 0;

  attr_waveArrayCount_read = new Tango::DevLong;
  *attr_waveArrayCount_read = 0;

  attr_nominalBits_read = new Tango::DevShort;
  *attr_nominalBits_read = 0;

  attr_horizontalInterval_read = new Tango::DevDouble;
  *attr_horizontalInterval_read = 0.;

  attr_horizontalOffset_read = new Tango::DevDouble;
  *attr_horizontalOffset_read = 0.;

  attr_verticalGain_read = new Tango::DevDouble;
  *attr_verticalGain_read = 0.;

  attr_verticalOffset_read = new Tango::DevDouble;
  *attr_verticalOffset_read = 0.;

  attr_triggerTime_read = new Tango::DevString;
  *attr_triggerTime_read = new char[MAX_STRING_LENGTH];
  ::memset(*attr_triggerTime_read, 0, MAX_STRING_LENGTH * sizeof(char));

  //- Initialise variables to default values
  //--------------------------------------------
  waveform_mgr_ = 0;
  channel_data_ = 0;
  data_length_  = 0;

  get_device_property();

  //---------------------------------------------------------
  //- instanciate the log_adapter & inner_appender in order to manage logs
  //---------------------------------------------------------
//   try
//   {
//     DEBUG_STREAM << "Create the YatLogAdapter in order to use logs from third party library." << endl;
//     yat4tango::YatLogAdapter::initialize(this);
// 
//     DEBUG_STREAM << "Create the InnerAppender in order to manage logs." << endl;
//     yat4tango::InnerAppender::initialize(this, 512);
//   }
//   catch (Tango::DevFailed& df)
//   {
//     ERROR_STREAM << df << endl;
//     set_state(Tango::FAULT);
//     set_status("OUT OF MEMORY : Failed to create log manager !");
//     return;
//   }

  char * ipAdd = 0;
  if (!waveform_mgr_)
  {
	//- transform string to char*
	std::size_t length = iPaddress.size();
	ipAdd = new char[length+1];
	for(std::size_t i=0; i<length; i++)
	  ipAdd[i] = iPaddress[i];
	ipAdd[length] = '\0';
	//- create the waveform obj
	waveform_mgr_ = new WaveformMgr(this, ipAdd);
  }
  //- prepare task
  if ( waveform_mgr_ )
  {
	//- add channel to monitor
	waveform_mgr_->add_channel(channelName);
	//- start thread (if not already started)
	if ( !waveform_mgr_->periodic_msg_enabled() )
	  waveform_mgr_->go();
  }
  //- release temporary allocation
  if (ipAdd)
  {
	delete [] ipAdd;
	ipAdd = 0;
  }
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::get_device_property()
//
// description :  Read the device properties from database.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::get_device_property()
{
  //- Initialize your default values here.
  //------------------------------------------
  iPaddress   = "172.16.35.1";
  channelName = "C2";

  //- Read device properties from database.(Automatic code generation)
  //-------------------------------------------------------------
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("IPaddress"));
	dev_prop.push_back(Tango::DbDatum("ChannelName"));

	//	Call database and extract values
	//--------------------------------------------
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_property(dev_prop);
	Tango::DbDatum	def_prop, cl_prop;
	AcquireWaveformLecroyClass	*ds_class =
		(static_cast<AcquireWaveformLecroyClass *>(get_device_class()));
	int	i = -1;

	//	Try to initialize IPaddress from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  iPaddress;
	else {
		//	Try to initialize IPaddress from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  iPaddress;
	}
	//	And try to extract IPaddress value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  iPaddress;

	//	Try to initialize ChannelName from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  channelName;
	else {
		//	Try to initialize ChannelName from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  channelName;
	}
	//	And try to extract ChannelName value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  channelName;

  //- End of Automatic code generation
  //-------------------------------------------------------------
  Tango::DbData data_put;
  if (dev_prop[0].is_empty())
  {
    Tango::DbDatum  propert("IPaddress");
    propert <<  iPaddress;
    data_put.push_back(propert);
  }
  if (dev_prop[1].is_empty())
  {
    Tango::DbDatum  propert("ChannelName");
    propert <<  channelName;
    data_put.push_back(propert);
  }

  //  End of Automatic code generation
  //------------------------------------------------------------------
  if( !data_put.empty() )
    get_db_device()->put_property(data_put);
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::always_executed_hook()
//
// description :  method always executed before any command is executed
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::always_executed_hook()
{

}

//+----------------------------------------------------------------------------
//
// method :   AcquireWaveformLecroy::read_attr_hardware()
//
// description :  Hardware acquisition for attributes.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_attr_hardware(vector<long> &attr_list)
{
  //- DEBUG_STREAM << "In read_attr_hardware for " << attr_list.size();
  //- DEBUG_STREAM << " attribute(s)" << endl;
yat::Timer t;

  //- check
  if ( !waveform_mgr_ )
  {
	FATAL_STREAM << "read_attr_hardware : device initialization failed!" << std::endl;
    Tango::Except::throw_exception("COMMUNICATION_BROKEN",
                                   "Device initialization failed : no Waveform manager.",
                                   "AcquireWaveformLecroy::read_attr_hardware()"
                                   );
  }
  //- get waveform data 
  try
  {
    channel_data_ = waveform_mgr_->get_channel_data();
	if (channel_data_)
	{
	  data_length_ = channel_data_->wave_array_count_;
	}
  }
  catch(const lecroy::WaveformException& we)
  {
    FATAL_STREAM << "R ATTR HW -> caught Lecroy exception!" << std::endl;
    Tango::DevFailed df = lecroy_to_tango_exception(we);
    FATAL_STREAM << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_BROKEN",
                                      "Cannot call get_waveform_data.",
                                      "AcquireWaveformLecroy::read_attr_hardware()",
                                      Tango::ERR
                                      );
  }
  catch(Tango::DevFailed& df)
  {
    FATAL_STREAM << "R ATTR HW -> caught DevFailed exception :\n" << df << std::endl;
    Tango::Except::re_throw_exception(df,
                                      "COMMUNICATION_BROKEN",
                                      "Failed to update channel data.",
                                      "AcquireWaveformLecroy::read_attr_hardware()",
                                      Tango::ERR
                                      );
  }
  catch(...)
  {
    FATAL_STREAM << "R ATTR HW -> caught [...] exception!" << std::endl;
    throw;
  }

  INFO_STREAM << "READ_HW : all data are now updated. DONE in " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_verticalScaledData
//
// description :  Extract real attribute values for verticalScaledData acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_verticalScaledData(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_verticalScaledData(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;

//   for(std::size_t idx = 0; idx < data_length_; idx++)
//     attr_verticalScaledData_read[idx] = channel_data_->vertical_scaled_data_[idx];
//   attr.set_value(attr_verticalScaledData_read, data_length_);

  if ( data_length_ )
	attr.set_value(&channel_data_->vertical_scaled_data_[0], data_length_);
  DEBUG_STREAM << "read_verticalScaledData : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_triggerTime
//
// description :  Extract real attribute values for triggerTime acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_triggerTime(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_triggerTime(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;

  //  Add your own code here
  std::string response = channel_data_->trigger_time_;
  strcpy(*attr_triggerTime_read, response.c_str());

  attr.set_value(attr_triggerTime_read);
  DEBUG_STREAM << "read_triggerTime : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_verticalOffset
//
// description :  Extract real attribute values for verticalOffset acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_verticalOffset(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_verticalOffset(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;

  //- Add your own code here
 *attr_verticalOffset_read = channel_data_->vertical_offset_;

  attr.set_value(attr_verticalOffset_read);
  DEBUG_STREAM << "read_verticalOffset : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_verticalGain
//
// description :  Extract real attribute values for verticalGain acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_verticalGain(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_verticalGain(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;

  //- Add your own code here
 *attr_verticalGain_read = channel_data_->vertical_gain_;

  attr.set_value(attr_verticalGain_read);
  DEBUG_STREAM << "read_verticalGain : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_horizontalOffset
//
// description :  Extract real attribute values for horizontalOffset acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_horizontalOffset(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_horizontalOffset(Tango::Attribute &attr) entering... "<< endl;
 yat::Timer t;

  //- Add your own code here
  *attr_horizontalOffset_read = channel_data_->horizontal_offset_;
  
  attr.set_value(attr_horizontalOffset_read);
  DEBUG_STREAM << "read_horizontalOffset : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_horizontalInterval
//
// description :  Extract real attribute values for horizontalInterval acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_horizontalInterval(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_horizontalInterval(Tango::Attribute &attr) entering... "<< endl;
 yat::Timer t;

  //- Add your own code here
  *attr_horizontalInterval_read = channel_data_->horizontal_interval_;
  
  attr.set_value(attr_horizontalInterval_read);
  DEBUG_STREAM << "read_horizontalInterval : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_nominalBits
//
// description :  Extract real attribute values for nominalBits acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_nominalBits(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_nominalBits(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;
  //- Add your own code here
  *attr_nominalBits_read = channel_data_->nominal_bits_;
  
  attr.set_value(attr_nominalBits_read);
  DEBUG_STREAM << "read_nominalBits : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_waveArrayCount
//
// description :  Extract real attribute values for waveArrayCount acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_waveArrayCount(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_waveArrayCount(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;
  //- Add your own code here
  *attr_waveArrayCount_read = data_length_;
  
  attr.set_value(attr_waveArrayCount_read);
  DEBUG_STREAM << "read_waveArrayCount : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_waveArray2
//
// description :  Extract real attribute values for waveArray2 acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_waveArray2(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_waveArray2(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;
  //  Add your own code here
  *attr_waveArray2_read = channel_data_->wave_array_2_;
  
  attr.set_value(attr_waveArray2_read);
  DEBUG_STREAM << "read_waveArray2 : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_waveArray1
//
// description :  Extract real attribute values for waveArray1 acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_waveArray1(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_waveArray1(Tango::Attribute &attr) entering... "<< endl;
yat::Timer t;
  //- Add your own code here
  *attr_waveArray1_read = channel_data_->wave_array_1_;
  
  attr.set_value(attr_waveArray1_read);
  DEBUG_STREAM << "read_waveArray1 : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method :     AcquireWaveformLecroy::read_rawWaveformData
//
// description :  Extract real attribute values for rawWaveformData acquisition result.
//
//-----------------------------------------------------------------------------
void AcquireWaveformLecroy::read_rawWaveformData(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "AcquireWaveformLecroy::read_rawWaveformData(Tango::Attribute &attr) entering... "<< endl;
  //- Add your own code here
  yat::Timer t;
  //- get waveform data
  //-for(std::size_t idx = 0; idx < data_length_; idx++)
  //-  attr_rawWaveformData_read[idx] = channel_data_->raw_waveform_data_[idx];
  //-attr.set_value(attr_rawWaveformData_read, data_length_);
  
  if ( data_length_ )
	attr.set_value(&channel_data_->raw_waveform_data_[0], data_length_);
  DEBUG_STREAM << "read_rawWaveformData : " << t.elapsed_msec() << " milliseconds.\n" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method : AcquireWaveformLecroy::lecroy_to_tango_exception()
//
// description :   Extract real attribute values from
//        hardware acquisition result.
//
//-----------------------------------------------------------------------------
Tango::DevFailed AcquireWaveformLecroy::lecroy_to_tango_exception(const lecroy::LecroyException& de)
{
  Tango::DevErrorList error_list(de.errors.size());
  error_list.length(de.errors.size());

  for(size_t i = 0; i < de.errors.size(); i++)
  {
    error_list[i].reason = CORBA::string_dup(de.errors[i].reason.c_str());
    error_list[i].desc   = CORBA::string_dup(de.errors[i].desc.c_str());
    error_list[i].origin = CORBA::string_dup(de.errors[i].origin.c_str());

    switch(de.errors[i].severity)
    {
    case lecroy::WARN:
      error_list[i].severity = Tango::WARN;
      break;
    case lecroy::PANIC:
      error_list[i].severity = Tango::PANIC;
      break;
    case lecroy::ERR:
    default:
      error_list[i].severity = Tango::ERR;
      break;
    }
  }

  return Tango::DevFailed(error_list);
}

//+------------------------------------------------------------------
/**
 *	method:	AcquireWaveformLecroy::write_read
 *
 *	description:	method to execute "WriteRead"
 *	Command to send a specific command to the Lecroy device
 *
 * @param	argin	command to send
 * @return	device response (if any)
 *
 */
//+------------------------------------------------------------------
Tango::DevString AcquireWaveformLecroy::write_read(Tango::DevString argin)
{
  //  POGO has generated a method core with argout allocation.
  //  If you would like to use a static reference without copying,
  //  See "TANGO Device Server Programmer's Manual"
  //    (chapter : Writing a TANGO DS / Exchanging data)
  //------------------------------------------------------------
  //- DEBUG_STREAM << "AcquireWaveformLecroy::write_read(): entering... !" << endl;
yat::Timer t;
  Tango::DevString argout;
  
  std::string resp = waveform_mgr_->write_read(argin);
  
  std::size_t length = resp.size();

  if ( length > MAX_STRING_LENGTH )
	length = MAX_STRING_LENGTH;
   
  argout = new char[length+1];
  ::strcpy(argout, resp.c_str());
  
  INFO_STREAM << "WRITE_READ DONE in " << t.elapsed_msec() << " milliseconds.\n" << std::endl;

  return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AcquireWaveformLecroy::dev_state
 *
 *	description:	method to execute "State"
 *	This command gets the device state (stored in its <i>device_state</i> data member) and returns it to the caller.
 *
 * @return	State Code
 *
 */
//+------------------------------------------------------------------
Tango::DevState AcquireWaveformLecroy::dev_state()
{
	Tango::DevState	argout = Tango::UNKNOWN;
	DEBUG_STREAM << "AcquireWaveformLecroy::dev_state(): entering... !" << endl;

	//	Add your own code to control device here
	if ( !waveform_mgr_ )
	{
	  argout = Tango::FAULT;
	  set_status("Initialization failed : no manager to gather channel data.");
	}
	
	if ( !waveform_mgr_->periodic_msg_enabled() )
	{
	  argout = Tango::ALARM;
	  set_status("Task is not started : cannot gather channel data.");
	}
	
	if ( !waveform_mgr_->is_connected() )
	{
	  argout = Tango::ALARM;
	  set_status("No socket connection with Lecroy.");
	}
	
	if ( !data_length_ )
	{
	  argout = Tango::ALARM;
	  set_status("Failed to gather channel data : no data.");
	}
	
	if ( argout != Tango::FAULT && argout != Tango::ALARM )
	{
	  errors_ = waveform_mgr_->get_errors();
	  if ( errors_.empty() )
	  {
		argout = Tango::ON;
		set_status("No error(s).");
	  }
	  else
	  {
		argout = Tango::ALARM;
		set_status(errors_.c_str());
	  }
	}
	
	set_state(argout);
	return argout;
}

}	//	namespace
